#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <SDL/SDL.h>

#ifndef RES_DIR
#define RES_DIR "images/"
#endif

// ===========================================================================
// Constantes.

const int MAX_MAP_SIZE = 10;
const int PAWN_SIZE = 20;
const int TILE_SIZE = 42;

const char* tileOnFile = RES_DIR "tileOn.bmp";
const char* tileOffFile = RES_DIR "tileOff.bmp";

const struct { int dl, dc; } DEP[4] = {
    {1, 0}, {-1, 0}, {0, 1}, {0, -1}
};

// ===========================================================================
// Prototypes des procdures SDL.

void loadSurfaces();
void freeSurfaces();
void redrawMap();

// ===========================================================================
// Donnes globales.

bool map[MAX_MAP_SIZE][MAX_MAP_SIZE];
int mapHeight, mapWidth;
int curLine, curCol;

int nbVisits[MAX_MAP_SIZE][MAX_MAP_SIZE];
char way[MAX_MAP_SIZE * MAX_MAP_SIZE];

SDL_Event event;
SDL_Surface *pawn;
SDL_Surface *screen;
SDL_Surface *tileOff, *tileOn;

// ===========================================================================
// Fonctions triviales.

inline bool clean()
{
    bool isClean = true;
    
    for (int l = 0; l < mapHeight; l++)
        for (int c = 0; c < mapWidth; c++)
            isClean &= map[l][c];
    
    return isClean;
}

inline void error(char *msg)
{
    fprintf(stderr, msg);
    fprintf(stderr, "\n");
    fprintf(stderr, "SDL_GetError() : %s\n", SDL_GetError());
    
    freeSurfaces();
    exit(-1);
}

inline void invert(int l, int c)
{
    map[l][c] = !map[l][c];
}

// ===========================================================================
// Fonctions de mise  jour des donnes.

int update(int l, int c)
{
    if (l < 0 || l >= mapHeight || c < 0 || c >= mapWidth)
        return 0;
    
    for (int v = 0; v < 4; v++) {
        int nl = l + DEP[v].dl;
        int nc = c + DEP[v].dc;
        if (nl >= 0 && nl < mapHeight && nc >= 0 && nc < mapWidth)
            invert(nl, nc);
    }
    
    nbVisits[l][c]++;
    invert(l, c);
    return 1;
}

void reset()
{
    for (int l = 0; l < mapHeight; l++) {
        for (int c = 0; c < mapWidth; c++) {
            nbVisits[l][c] = 0;
            map[l][c] = false;
        }
    }
    
    way[0] = 0;
    curCol = 0;
    curLine = mapHeight - 1;
    update(curLine, curCol);
}

void writeSolution()
{
    char fstr[42];
    sprintf(fstr, "grille-%02dx%02d.txt", mapWidth, mapHeight);
    freopen(fstr, "w", stdout);
    
    for (int l = 0; l < mapHeight; l++) {
        for (int c = 0; c < mapWidth; c++)
            putchar(nbVisits[l][c] % 2 == 0 ? '_' : 'x');
        putchar('\n');
    }
}

// ===========================================================================
// Fonctions d'affichage.

void loadSurfaces()
{
    int width = TILE_SIZE * mapWidth;
    int height = TILE_SIZE * mapHeight;
    int flags = SDL_SWSURFACE | SDL_DOUBLEBUF;
    
    screen = SDL_SetVideoMode(width, height, 32, flags);
    
    if (!screen)
        error("Impossible d'initialiser le mode video.");
    
    pawn = SDL_CreateRGBSurface(SDL_SWSURFACE,
         PAWN_SIZE, PAWN_SIZE, 32, 0, 0, 0, 0);
    SDL_FillRect(pawn, NULL, SDL_MapRGB(screen->format, 100, 0, 0));
    
    tileOn = SDL_LoadBMP(tileOnFile);
    tileOff = SDL_LoadBMP(tileOffFile);
    
    if (!pawn || !screen || !tileOn || !tileOff)
        error("Erreur lors du chargement des surfaces.");
}

void freeSurfaces()
{
    SDL_FreeSurface(pawn);
    SDL_FreeSurface(tileOff);
    SDL_FreeSurface(tileOn);
    SDL_FreeSurface(screen);
}

void drawPawn()
{
    SDL_Rect rect;
    
    rect.w = PAWN_SIZE;
    rect.h = PAWN_SIZE;
    rect.x = TILE_SIZE * curCol + 11;
    rect.y = TILE_SIZE * curLine + 11;
    
    SDL_BlitSurface(pawn, NULL, screen, &rect);
}

void drawMap()
{
    SDL_Rect rect;
    SDL_Surface *tile;
    
    rect.w = TILE_SIZE;
    rect.h = TILE_SIZE;
    
    for (int l = 0; l < mapHeight; l++) {
        for (int c = 0; c < mapWidth; c++) {
            rect.x = TILE_SIZE * c;
            rect.y = TILE_SIZE * l;
            tile = (map[l][c]) ? tileOn : tileOff;
            SDL_BlitSurface(tile, NULL, screen, &rect);
        }
    }
}

inline void updateScreen()
{
    SDL_Flip(screen);
}

// ===========================================================================
// Fonctions de gestion des commandes.

void parseKeydown()
{
    switch (event.key.keysym.sym) {
        case SDLK_LEFT:
            curCol -= update(curLine, curCol - 1);
            strcat(way, "E");
            break;
        case SDLK_RIGHT:
            curCol += update(curLine, curCol + 1);
            strcat(way, "O");
            break;
        case SDLK_UP:
            curLine -= update(curLine - 1, curCol);
            strcat(way, "N");
            break;
        case SDLK_DOWN:
            curLine += update(curLine + 1, curCol);
            strcat(way, "S");
            break;
        case SDLK_RETURN:
            reset();
            break;
        case SDLK_ESCAPE:
            error("Interruption.");
            break;
        default:
            break;
    }
}

void waitForExit()
{
    SDL_WaitEvent(&event);
    
    if (event.type == SDL_QUIT)
        return;
    else if (event.type == SDL_KEYDOWN)
        switch (event.key.keysym.sym) {
            case SDLK_ESCAPE:
                return;
            case SDLK_RETURN:
                return;
            case SDLK_w:
                writeSolution();
                return;
            default:
                break;
        }
    
    waitForExit();
}

// ===========================================================================
// Fonction principale.

int main(int argc, char *argv[])
{
    if (argc >= 3) {
        mapHeight = atoi(argv[2]);
        mapWidth = atoi(argv[1]);
    } else {
        srand(time(0));
        mapHeight = (rand() % (MAX_MAP_SIZE - 1)) + 1;
        mapWidth = (rand() % (MAX_MAP_SIZE - 1)) + 1;
    }
    
    if (mapWidth < 0 || mapWidth > MAX_MAP_SIZE ||
        mapHeight < 0 || mapHeight > MAX_MAP_SIZE)
        error("Dimensions de grille invalides.");
    
    reset();
    
    if (SDL_Init (SDL_INIT_VIDEO) < 0)
        error("Impossible d'initialiser SDL.");
    
    atexit(SDL_Quit);
    loadSurfaces();
    SDL_WM_SetCaption("Binary Walk", NULL);
    
    while (!clean()) {
        drawMap();
        drawPawn();
        updateScreen();
        SDL_WaitEvent(&event);
        if (event.type == SDL_QUIT)
            break;
        else if (event.type == SDL_KEYDOWN)
            parseKeydown();                
    }
    
    drawMap();
    updateScreen();
    freeSurfaces();
    waitForExit();
    return 0;
}
